### LDA for PsychTopics App ###
# André Bittermann, ZPID, Trier
# abi@leibniz-psychology.org



load("DATENSATZ.RData") # data frame with publications in rows and features in columns

library(quanteda)
library(topicmodels)


### preprocessing ----

data_corpus <- corpus(CT.lda)

DFM <- data_corpus %>% 
  tokens() %>% 
  tokens_tolower() %>% 
  tokens_remove(pattern = stopwords) %>%
  tokens(remove_punct = FALSE, remove_numbers = FALSE, remove_symbols = FALSE, remove_separators = TRUE) %>% dfm()

DTM <- convert(DFM, to = "topicmodels")



### alpha ----

alphalist <- c(0.0001, 0.0005, 0.001) 
seedlist <- c(12, 4321, 28, 3323) # 4 different random seeds are used.
cores <- 4 # parallel computing
lda_besta <- list() # saves the best models across all seeds.
CO <- list()
EX <- list()
COEX <- list()
COEX_list <- list()

library(topicmodels)
library(quantqual) # devtools::install_github("AndreasFischer1985/quantqual")
source("topic_coherence.R") # by A. Niekler & G. Wiedemann upon request, https://tm4ss.github.io/

for (i in 1:length(alphalist)){
  a <- alphalist[i]
  cluster <- parallel::makeCluster(cores)
  parallel::setDefaultCluster(cluster)
  parallel::clusterExport(varlist = c("DTM.sample", "a"), envir = environment())
  lda_temp <- parallel::parLapply(X = seedlist, fun = function(x) {
    topicmodels::LDA(DTM.sample, k = 400, method = "Gibbs", control = list(alpha=a, delta=0.01, nstart=1, seed=x))
  })
  parallel::stopCluster(cluster)
  
  for (j in 1:length(seedlist)) {
    EX[[j]] <- mean(quantqual::exclusivityLDA(lda_temp[[j]], as.matrix(DTM.sample), num.terms = 100))
    CO[[j]] <- mean(tmca_topic_coherence(DTM.sample, posterior(lda_temp[[j]])$terms, N = 100))
  }
  
  COEX_list[[i]] <- rowMeans(cbind(scale(unlist(EX)), scale(unlist(CO)))) # mean coherence and exclusivity (scaled)
  COEX[[i]] <- max(unlist(COEX_list[[i]])) # value of max COEX
  lda_besta[[i]] <- lda_temp[[which.max(unlist(COEX_list[[i]]))]] # save the best model
}


doc <- 283 # check various documents
par(mfrow=c(3,1))
for (i in 1:length(alphalist)){
  plot(posterior(lda_besta[[i]])$topics[doc,], type = "h", 
       main = paste("Theta-Verteilung für Dokument", doc, "bei alpha =", alphalist[i], 
                    "Max_theta =", round(max(posterior(lda_besta[[i]])$topics[doc,]), 4)), 
       ylab = "Theta", xlab = "Topic")
}


# compare document content with topic content

sample[doc, 6] # doc content

a1_topics <- which(posterior(lda_besta[[1]])$topics[doc,] > 0.05) # get topic numbers
a2_topics <- which(posterior(lda_besta[[2]])$topics[doc,] > 0.05)
a3_topics <- which(posterior(lda_besta[[3]])$topics[doc,] > 0.05)

as.matrix(terms(lda_besta[[1]], 15))[,a1_topics] # topic contents
as.matrix(terms(lda_besta[[2]], 15))[,a2_topics]
as.matrix(terms(lda_besta[[3]], 15))[,a3_topics]

# alpha = 0.005 works best



### model selection ----


klist <- seq(250, 550, by = 25)
seedlist <- c(12, 4321, 28, 3323) # 4 different random seeds are used.
cores <- 4 # parallel computing
lda_bestk <- list() # saves the best models across all seeds. Best is defined using the mean of Semantic Coherence and Exclusivity.
CO <- list()
EX <- list()
COEX <- list()
COEX_list <- list()

for (i in 1:length(klist)){
  a <- klist[i]
  cluster <- parallel::makeCluster(cores)
  parallel::setDefaultCluster(cluster)
  parallel::clusterExport(varlist = c("DTM", "a"), envir = environment())
  lda_temp <- parallel::parLapply(X = seedlist, fun = function(x) {
    topicmodels::LDA(DTM, k = a, method = "Gibbs", control = list(alpha=0.005, delta=0.01, nstart=1, seed=x))
  })
  parallel::stopCluster(cluster)
  
  for (j in 1:length(seedlist)) {
    EX[[j]] <- mean(quantqual::exclusivityLDA(lda_temp[[j]], as.matrix(DTM), num.terms = 100))
    CO[[j]] <- mean(tmca_topic_coherence(DTM, posterior(lda_temp[[j]])$terms, N = 100))
  }
  
  COEX_list[[i]] <- rowMeans(cbind(scale(unlist(EX)), scale(unlist(CO)))) # mean coherence and exclusivity (scaled)
  COEX[[i]] <- max(unlist(COEX_list[[i]])) # value of max COEX
  lda_bestk[[i]] <- lda_temp[[which.max(unlist(COEX_list[[i]]))]] # save the best model
}



### metrics ----

## evaluation metrics
LL_best <- list()
EX_best <- list()
CO_best <- list()


for (i in 1:length(klist)){
  LL_best[[i]] <- lda_bestk[[i]]@loglikelihood
  EX_best[[i]] <- mean(exclusivityLDA(lda_bestk[[i]], as.matrix(DTM), num.terms = 100))
  CO_best[[i]] <- mean(tmca_topic_coherence(DTM, posterior(lda_bestk[[i]])$terms, N = 100))
}

plot(scale(unlist(CO_best)), type = "l", col = "black", xaxt = "n", ylim = c(-1.5, 1.5), lwd = 2,
     main = "LDA Model Evaluation", ylab = paste("Scaled Score"), 
     xlab = "Number of Topics")
lines(scale(unlist(EX_best)), type = "l", col = "blue", lwd = 2)
lines(scale(unlist(LL_best)), type = "l", col = "purple", lwd = 2)
axis(1, at=1:length(klist), labels=klist)
legend("bottomright", c("Semantic Coherence", "Exclusivity", "Likelihood"), 
       col = c("black", "blue", "purple"), lty = "solid", lwd = 2)

# indicates k = 325


### theta inspections ----

doc <- 28 # pick various documents
par(mfrow=c(3,5))
for (i in 1:length(klist)){
  plot(posterior(lda_bestk[[i]])$topics[doc,], type = "h", 
       main = paste("Doc", doc, "bei k =", klist[i], 
                    "Max_theta =", round(max(posterior(lda_bestk[[i]])$topics[doc,]), 4)), 
       ylab = "Theta", xlab = "Topic")
}

# k = 325 works best



### final model ----

LDA.shiny <- lda_bestk[[which(klist == 325)]]



### topic reliability ----

# Repeat inference with different seeds.
# These models will be compared pairwise in order to determine model reliability.
# Which topics occur across different inference runs (with different seeds)?

# different seeds
seedlist_rel <- c(127, 47321, 278, 37323, 2721, 127722, 97039, 173, 317231, 98778) 
cluster <- parallel::makeCluster(cores)
parallel::setDefaultCluster(cluster)
parallel::clusterExport(varlist = c("DTM", "a"), envir = environment())
lda_temp <- parallel::parLapply(X = seedlist_rel, fun = function(x) {
  topicmodels::LDA(DTM, k = 325, method = "Gibbs", control = list(alpha=0.005, delta=0.01, nstart=1, seed=x))
})
parallel::stopCluster(cluster)


source("reliability.R") # by A. Niekler & G. Wiedemann
# Niekler, A.: Automatisierte Verfahren für die Themenanalyse nachrichtenorientierter Textquellen: 
# Dissertation zur Erlangung des akademischen Grades Doktor-Ingenieur (Dr.-Ing.) im Fachgebiet Informatik, 
# http://asv.informatik.uni-leipzig.de/publication/file/350/Niekler_Diss.pdf, (2016).


topic_rel <- topic # df with topic numbers and terms

for (i in 1:length(seedlist_rel)){
  tm_fit1 <- fromTopicModelsPackage(LDA.shiny)
  tm_fit2 <- fromTopicModelsPackage(lda_temp[[i]])
  
  relCosine <- reliability_cosine(tm_fit1, tm_fit2, topWordsToMatch = 20, threshold = 0.8)
  #print(relCosine$reliability)
  topicmatches <- relCosine$ids
  #sort(topicmatches[,1]) # topic numbers (in LDA_final) that occur across inference runs
  
  rel_df <- data.frame(topicmatches[topicmatches[,1] > 0, 1], rep(1, length(topicmatches[topicmatches[,1] > 0, 1])))
  names(rel_df) <- c("Nr..", paste0("rel_",i))
  topic_rel <- dplyr::left_join(topic_rel, rel_df, by = "Nr..")
}

topic_rel$rel_sum <- rowSums(topic_rel[,4:length(topic_rel)], na.rm = TRUE)
table(topic_rel$rel_sum)

keep <- topic_rel[topic_rel$rel_sum >= 8, 1] # numbers of topics that occured in at least 80 % of other inference runs



### keep only reliable topics ----

k <- length(keep)

lda.theta <- posterior(LDA.shiny)$topics[, keep]
lda.topterms <- as.matrix(terms(LDA.shiny, 10))[, keep]

dim(lda.theta)
dim(lda.topterms)
lda.theta[1:5,1:5]
lda.topterms[,1:5]


# topic labels
top5termsPerTopicProb <- terms(LDA.shiny, 5)[, keep]
topicNames <- apply(top5termsPerTopicProb, 2, paste, collapse=", ")



### data frame with topics, theta, and labels ----

topic <- data.frame(
  TopicNumber = 1:k,
  TopicProportions = colMeans(lda.theta))
topic$terms <- topicNames



### theta objects ----

n.by.PY <- t(table(DATENSATZ$PY))
rownames(n.by.PY) <- "Frequency"
n.by.PY

years <- levels(factor(DATENSATZ$PY)) # list of years

### time.series original code by M. Ponweiser: https://github.com/mponweiser/thesis-LDA-in-R

theta_mean_by_year_by <- by(lda.theta, DATENSATZ$PY, colMeans)
theta_mean_by_year <- do.call("rbind",theta_mean_by_year_by)
colnames(theta_mean_by_year) = paste(1:k)
theta_mean_by_year_ts <- ts(theta_mean_by_year, start = as.integer(years[1])) # Time-series
theta_mean_by_year_time <- time(theta_mean_by_year)

# theta by year with labels
theta_year <- theta_mean_by_year
colnames(theta_year) <- topicNames

# Topic Overview
topic[ ,c(1,2,3)] <- topic[ ,c(1,3,2)] # reorder
names(topic) <- c("Nr..", "Thema", "Prävalenz") # renamed in app.R



### LDA beta probabilities ----

lda.beta <- posterior(LDA.shiny)$terms[keep, ]

## tidy beta
# original code by Silge & Robinson (2017)
# Silge, J. & Robinson, D. (2017). Text Mining with R. A Tidy Approach.
# http://tidytextmining.com/topicmodeling.html

library(tidytext)
library(dplyr)

# term probabilities by topic
ap_word <- tidytext::tidy(LDA.shiny, matrix = "beta")
head(ap_word)

# first 5 words of topic
ap_top_terms <- ap_word %>%
  group_by(topic) %>%
  top_n(5, beta) %>%
  ungroup() %>%
  arrange(topic, -beta)


ap <- aggregate(beta ~ topic, data = ap_top_terms, paste, collapse = ", ")[keep, ]
ap$topic <- 1:k


## create boosting factors

# https://stackoverflow.com/questions/4350440/split-data-frame-string-column-into-multiple-columns
library(data.table)
setDT(ap)[, paste0("beta", 1:6) := tstrsplit(beta, ", ")]

ap <- ap[,c(1,3:7)]

ap$b1 <- as.numeric(ap$beta1) / as.numeric(ap$beta5)
ap$b2 <- as.numeric(ap$beta2) / as.numeric(ap$beta5)
ap$b3 <- as.numeric(ap$beta3) / as.numeric(ap$beta5)
ap$b4 <- as.numeric(ap$beta4) / as.numeric(ap$beta5)

booster.lda <- as.data.frame(t(ap[,c(7:10)]))
rownames(booster.lda) <- NULL
colnames(booster.lda) <- 1:k

booster <- round(booster.lda, 2)



### save files ----

saveRDS(theta_year, file = "./data/theta_year.rds")
saveRDS(theta_mean_by_year, file = "./data/theta_mean_by_year.rds")
saveRDS(theta_mean_by_year_time, file = "./data/theta_mean_by_year_time.rds")
saveRDS(theta_mean_by_year_ts, file = "./data/theta_mean_by_year_ts.rds")
saveRDS(years, file = "./data/years.rds")
saveRDS(topic, file = "./data/topic.rds")
saveRDS(booster, file = "./data/booster.rds")